/*
  ip_frag.c

  Dug Song <dugsong@anzen.com>

  Copyright (c) 1999 Anzen Computing. All rights reserved.

  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions
  are met:
  
  1. Redistributions of source code must retain the above copyright
     notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  3. All advertising materials mentioning features or use of this software
     must display the following acknowledgement:
     This product includes software developed by Anzen Computing.
  4. Neither the name of Anzen Computing nor the names of its
     contributors may be used to endorse or promote products derived
     from this software without specific prior written permission.

  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
  IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
  DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
  DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
  GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER
  IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
  OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
  ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

  $Id: ip_frag.c,v 1.1 1999/05/18 15:34:05 dugsong Exp $
*/

#include <libnet.h>
#include "list.h"
#include "ip_frag.h"

int ip_frag_preserve_headers = 0;

void
ip_frag_init(int preserve_headers)
{
  ip_frag_preserve_headers = preserve_headers;
}

ELEM *
ip_frag_make(u_char *pkt, int pktlen, int fragsize)
{
  ELEM *new, *list = NULL;
  struct ip *iph = (struct ip *)pkt;
  int len, ip_hl = iph->ip_hl * 4;
  u_short ip_len = ntohs(iph->ip_len);
  u_char *ip_blob = pkt + ip_hl;
  u_char *ip_end = pkt + ip_len;
  u_char *p, *data;

  /* Make sure fragmentation is valid for this packet. */
  if (pktlen <= ip_hl + fragsize || fragsize % 8 != 0)
    return NULL;
  
  /* Fragment packet. */
  for (p = ip_blob ; p < ip_end ; p += len) {
    if ((p == ip_blob) && ip_frag_preserve_headers) {
      if (ip_end - p <= 24) return NULL;
      len = 24;
    }
    else
#ifdef BSD
    /* XXX - weird OpenBSD sendto() bug: < 4 byte payload returns ENOBUFS? */
    if ((ip_end - p) <= (fragsize + 4))
      len = ip_end - p;
    else
#endif /* BSD */
    len = MIN(ip_end - p, fragsize);

    if (libnet_init_packet(ip_hl + len, &data) == -1)
      return NULL;

    /* Copy in IP header and data. */
    memcpy(data, pkt, ip_hl);
    memcpy(data + ip_hl, p, len);

    /* Correct IP length, IP fragment offset. */
    ((struct ip *)data)->ip_len = htons(ip_hl + len);
    ((struct ip *)data)->ip_off =
      htons(((p + len < ip_end) ? IP_MF : 0) | ((p - ip_blob) >> 3));

    /* Add to our IP fragment list. */
    new = list_elem(data, ip_hl + len);
    
    if (!(list = list_add(list, new)))
      return NULL;
  }
  return (list->head);
}

ELEM *
ip_frag_add_overwrite(ELEM *list)
{
  ELEM *f, *new;
  u_char nulls[PACKET], *newdata;
  int newoff, newlen;

  memset(nulls, 0, sizeof(nulls));
  
  for (f = list ; f && f->next ; f = f->next) {
    struct ip *iph = (struct ip *)f->data;
    int ip_hl = iph->ip_hl * 4;

    /* Build new null data fragment, half the length of the current fragment
       and overlapping its latter half. */
    newlen = (ntohs(iph->ip_len) - ip_hl) / 2;
    newoff = ntohs(iph->ip_off) + (newlen >> 3);

    if (libnet_init_packet(ip_hl + newlen, &newdata) == -1)
      return NULL;

    /* Copy in IP header and data. */
    memcpy(newdata, f->data, ip_hl);
    memcpy(newdata + ip_hl, nulls, newlen);

    /* Correct IP length, IP fragment offset. */
    ((struct ip *)newdata)->ip_len = htons(ip_hl + newlen);
    ((struct ip *)newdata)->ip_off = htons(newoff);

    /* Add null fragment after current fragment. */
    new = list_elem(newdata, IP_H + newlen);
    
    if (!list_add(f, new))
      return NULL;
    
    /* Swap them. */
    if (!list_swap(f))
      return NULL;
  }
  return list;
}
